"""
Plotting module for PyTorch prototyping

:author: Julian M. Kleber
"""
"""
Plotting module for PyTorch prototyping

:author: Julian M. Kleber
"""


from typing import Type, Optional, Dict, Any, List, Tuple
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from amarium.utils import load_json_from_file, prepare_file_name_saving
from collections import ChainMap


def plot_range_band_multi(
    result: List[Dict[str, float]],
    key_vals: List[str],
    file_name: str,
    save_dir: Optional[str] = None,
) -> None:
    fig, ax = plt.subplots()
    ax.set_xlabel("Training step")

    for i in range(len(key_vals)):
        max_val: List[float]
        min_val: List[float]
        avg_val: List[float]

        max_val, min_val, avg_val = unpack_cross_validation(
            result=result, key_val=key_vals[i]
        )

        ax.plot(avg_val, "-", label=key_vals[i])
        ax.set_ylim(0, 1.1)
        ax.fill_between(min_val, max_val, alpha=0.2)

    ax.set_ylabel("Value")
    ax.legend()

    save_publication_graphic(fig_object=fig, file_name=file_name, prefix=save_dir)


def plot_range_band_single(
    result: List[Dict[str, float]],
    key_val: str,
    file_name: str,
    save_dir: Optional[str] = None,
) -> None:
    """
    The plot_range_band function takes in a list of dictionaries, each dictionary containing the
    results from one run.
    It then plots the average value for each key_val (e.g., 'accuracy') and also plots a
    range band between the minimum and maximum values for that key_val across all runs.

    :param result:List[Dict[str: Used to Plot the results of each run.
    :param float]]: Used to Specify the type of data that is being passed into the function.
    :param key_val:str: Used to Specify which key in the dictionary to plot.
    :param file_name:str: Used to Save the plot as a png file.
    :return: A plot with the average value of a list, and the minimum and maximum values.

    :doc-author: Julian M. kleber
    """
    max_val: List[float]
    min_val: List[float]
    avg_val: List[float]

    max_val, min_val, avg_val = unpack_cross_validation(result=result, key_val=key_val)

    fig, ax = plt.subplots()
    ax.plot(avg_val, "-")
    ax.set_ylim(0, 1.1)
    ax.set_ylabel(key_val)
    ax.set_xlabel("Training step")
    ax.fill_between(min_val, max_val, alpha=0.8)

    save_publication_graphic(fig_object=fig, file_name=file_name, prefix=save_dir)


def unpack_cross_validation(
    result: List[Dict[str, float]], key_val: str
) -> Tuple[List[float]]:
    """
    The unpack_cross_validation function takes in a list of dictionaries, and a key value.
    It then unpacks the values associated with that key into three lists: max_val, min_val, avg_val.
    These lists are returned as a tuple.

    :param result:List[Dict[str: Used to Store the result of each iteration.
    :param float]]: Used to Store the results of the cross validation.
    :param key_val:str: Used to Specify which key in the dictionary to use for the unpacking.
    :return: The max, min and average value of the key_val parameter.

    :doc-author: Trelent
    """

    max_val = []
    min_val = []
    avg_val = []

    arr_res = np.zeros((len(result), len(result[0][key_val])))
    print(arr_res.shape)

    for i, res in enumerate(result):
        arr_res[i, :] = res[key_val]

    for i in range(arr_res.shape[1]):
        max_val.append(get_min(arr_res[:, i].tolist()))
        min_val.append(get_max(arr_res[:, i].tolist()))
        avg_val.append(get_avg(arr_res[:, i].tolist()))

    return (max_val, min_val, avg_val)


def save_publication_graphic(
    fig_object: Type[plt.figure], file_name: str, prefix: Optional[str] = None
) -> None:
    """
    The save_publication_graphic function saves the current figure to a file.

    The save_publication_graphic function saves the current figure to a file, with
    a default resolution of 300 dpi. The function also tightens up the layout of
    the plot before saving it, so that there is no wasted space around it in its saved form.

    :param file_name:str: Used to Specify the name of the file to be saved.
    :param prefix:Optional[str]=None: Used to Specify the directory where the file is saved.
    :return: None.

    :doc-author: Julian M. Kleber
    """

    file_name = prepare_file_name_saving(
        file_name=file_name, prefix=prefix, suffix=".png"
    )
    plt.tight_layout()
    plt.savefig(file_name, dpi=300)


def get_stacked_list(
    path_to_directory: str, column_name: str, num_cv: int, json_name: str
) -> List[Dict[str, float]]:
    results = []

    for i in range(num_cv):
        print(f"Attempting cv {i}")

        result = parse_acc_list_json(
            path_to_json=path_to_directory + f"CV_{i}/" + json_name
        )
        results.append(result)
        print(f"parsed results for CV {i}")

    return results


def parse_old_file_format_for_plot(
    stacked_list: List[float], path_to_json: str
) -> List[List[float]]:
    """
    The parse_old_file_format_for_plot function takes in a path to a json file and returns the following:

    :param path_to_json:str: Used to Specify the path to the json file that we want to parse.
    :return: A list of dictionaries.

    :doc-author: Julian M. Kleber
    """

    result_acc = parse_acc_list_json(path_to_json=path_to_json)
    train_frames_acc = parse_min_max_avg(result_acc["Train_acc"])
    test_frames_acc = parse_min_max_avg(result_acc["Test_acc"])
    return train_frames_acc, test_frames_acc


def parse_acc_list_json(path_to_json: str) -> Dict[str, Any]:
    """
    The parse_acc_list_json function takes in a path to a json file and returns the contents of that json file as a dictionary.
    The function also parses the "Acc" key in the dictionary, which contains lists of tuples containing train and test accuracy values.
    The function then separates these tuples into two separate lists, one for train accuracy values and one for test accuracy values.
    These new lists are added to the original dictionary under keys "Acc_train" and "Acc_val", respectively.

    :param path_to_json:str: Used to Specify the path to the json file.
    :return: A dictionary with the following keys:.

    :doc-author: Trelent
    """

    result = load_json_from_file(path_to_json)
    return result


def parse_acc_list_json_old_format(path_to_json: str) -> Dict[str, Any]:
    """
    The parse_acc_list_json function takes in a path to a json file and returns the contents of that json file as a dictionary.
    The function also parses the "Acc" key in the dictionary, which contains lists of tuples containing train and test accuracy values.
    The function then separates these tuples into two separate lists, one for train accuracy values and one for test accuracy values.
    These new lists are added to the original dictionary under keys "Acc_train" and "Acc_val", respectively.

    :param path_to_json:str: Used to Specify the path to the json file.
    :return: A dictionary with the following keys:.

    :doc-author: Trelent
    """

    result = load_json_from_file(path_to_json)
    acc_store = result["Acc"]
    print(len(acc_store))
    train_acc = []
    test_acc = []

    for i in range(len(acc_store)):
        train, test = acc_store[i]
        train_acc.append(train)
        test_acc.append(test)

    del result["Acc"]
    print(f"Len of train, {len(train_acc)}")
    result["Acc_train"] = train_acc
    result["Acc_val"] = test_acc

    return result


def parse_min_max_avg(result_list: List[List[float]]) -> List[float]:
    """
    The parse_min function takes a list of lists and returns the minimum value for each sublist.

    :param result_list:List[List[float]]: Used to Store the results of the simulation.
    :return: A list of the minimum values for each step.

    :doc-author: Trelent
    """

    minima = []
    maxima = []
    averages = []
    for i in range(len(result_list[0])):
        step_list = result_list[:, i]
        minimum = get_min(step_list=step_list)
        maximum = get_max(step_list=step_list)
        average = get_avg(step_list=step_list)
        minima.append(minimum)
        maxima.append(maximum)
        averages.append(average)
    return [minima, maxima, averages]


def get_avg(step_list: List[float]) -> float:
    """
    The get_avg function takes a list of floats and returns the average value.


    :param step_list:List[float]: Used to Tell the function that it will be taking a list of floats as an argument.
    :return: The mean of the step_list.

    :doc-author: Julian M. Kleber
    """

    return float(np.mean(step_list))


def get_max(step_list: List[float]) -> float:
    """
    The get_max function takes a list of floats and returns the maximum value in that list.


    :param step_list:List[float]: Used to Tell the function that step_list is a list of floats.
    :return: The maximum value in the list.

    :doc-author: Trelent
    """

    return float(np.max(step_list))


def get_min(step_list: List[float]) -> float:
    """
    The get_min function takes a list of floats and returns the minimum value in that list.

    :param step_list:List[float]: Used to Specify the type of parameter that is being passed into the function.
    :return: The minimum value in the step_list.

    :doc-author: Trelent
    """

    return float(np.min(step_list))



"""
Tests the file converter that converts a config file to a json for using to initialize a Run object.

"""

from carate.utils.convert_to_json import convert_py_to_json, read_file


def test_converter():
    ref_py = [
        "dataset_name = ZINC_test",
        "num_classes = 1",
        "num_features = 18",
        "model = CGCR",
        "device = torch.device(cuda if torch.cuda.is_available() else cpu)",
        "optimizer = None  # defaults to adams optimizer",
        "net_dimension = 364",
        "learning_rate = 0.0005",
        "dataset_save_path = data/",
        "test_ratio = 20",
        "batch_size = 64",
        "shuffle = True",
        "num_epoch = 2",
        "num_cv = 2",
        "result_save_dir = results/",
    ]
    ref_json = {
        "dataset_name": "ZINC_test",
        "num_classes": "1",
        "num_features": "18",
        "model": "CGCR",
        "device": "torch.device(cuda",
        "optimizer": None,
        "net_dimension": "364",
        "learning_rate": "0.0005",
        "dataset_save_path": "data/",
        "test_ratio": "20",
        "batch_size": "64",
        "shuffle": "True",
        "num_epoch": "2",
        "num_cv": "2",
        "result_save_dir": "results/",
    }

    input_file = "tests/config/config.py"
    output_file = "tests/config//config.json"

    raw_input_lines = read_file(file_name=input_file)  #
    assert raw_input_lines == ref_py

    result = convert_py_to_json(file_name=input_file, out_name=output_file)
    assert result == ref_json



"""Test module to test the Config, Runner, Dataset pattern
The initialization is a bit weird and needs proper testing leaving no 
doubt 

:author: Julian M. Kleber
"""
from carate.load_data import StandardDatasetTUDataset, StandardDatasetMoleculeNet


def test_tu_data_set():
    dataset_save_path = "tests/data/"
    dataset_name = "ZINC_test"
    test_ratio = 10
    batch_size = 64
    shuffle = True

    loader = StandardDatasetTUDataset(
        dataset_save_path=dataset_save_path,
        dataset_name=dataset_name,
        test_ratio=test_ratio,
        batch_size=batch_size,
    )

    test_dataset, train_dataset, test_loader, train_loader, dataset = loader.load_data(
        dataset_name, test_ratio, dataset_save_path
    )

    assert str(loader) == "StandardTUDataset"


def test_molnet_data_set():
    dataset_save_path = "tests/data/"
    dataset_name = "sider"
    test_ratio = 10
    batch_size = 64
    shuffle = True

    loader = StandardDatasetMoleculeNet(
        dataset_save_path=dataset_save_path,
        dataset_name=dataset_name,
        test_ratio=test_ratio,
        batch_size=batch_size,
    )

    test_dataset, train_dataset, test_loader, train_loader, dataset = loader.load_data(
        dataset_name, test_ratio, dataset_save_path
    )

    assert str(loader) == "StandardMoleculeNet"


def test_custom_size():
    dataset_save_path = "tests/data/"
    dataset_name = "sider"
    test_ratio = 10
    batch_size = 64
    shuffle = True
    custom_size = 100

    loader = StandardDatasetMoleculeNet(
        dataset_save_path=dataset_save_path,
        dataset_name=dataset_name,
        test_ratio=test_ratio,
        batch_size=batch_size,
        custom_size=custom_size,
    )

    test_dataset, train_dataset, test_loader, train_loader, dataset = loader.load_data(
        dataset_name, test_ratio, dataset_save_path, custom_size = custom_size
    )

    assert str(loader) == "StandardMoleculeNet"
    assert len(test_dataset) == 10
    assert len(train_dataset) == 90
    
    dataset_save_path = "tests/data/"
    dataset_name = "ZINC_test"
    test_ratio = 10
    batch_size = 64
    shuffle = True
    custom_size = 100 

    loader = StandardDatasetTUDataset(
        dataset_save_path=dataset_save_path,
        dataset_name=dataset_name,
        test_ratio=test_ratio,
        batch_size=batch_size,
    )

    test_dataset, train_dataset, test_loader, train_loader, dataset = loader.load_data(
        dataset_name, test_ratio, dataset_save_path, custom_size = custom_size
    )

    assert str(loader) == "StandardTUDataset"
    assert len(test_dataset) == 10
    assert len(train_dataset) == 90


"""
Test to verify if the initialization of the config file is correct.

:author: Julian M. Kleber
"""
import torch
from carate.load_data import StandardDatasetTUDataset

from carate.config import Config, ConfigInitializer
from carate.models import cgc_classification
from carate.evaluation import classification
from typing import Type


def test_config():
    config_filepath = "tests/config/classification_test_config_override.py"
    config = ConfigInitializer.from_file(file_name=config_filepath)
    assert config.dataset_name == "ENZYMES"
    assert config.num_classes == 6
    assert config.num_features == 3
    assert config.model == cgc_classification
    assert config.Evaluation.__repr__() == "Classification Evaluation Object"
    assert config.optimizer == "adams"
    assert config.net_dimension == 364
    assert config.learning_rate == 0.0005
    assert config.dataset_save_path == "tests/data/"
    assert config.test_ratio == 10
    assert config.batch_size == 64
    assert config.shuffle == True
    assert config.num_epoch == 2
    assert config.num_cv == 2
    assert config.device == torch.device("cuda" if torch.cuda.is_available() else "cpu")
    assert config.result_save_dir == "tests/results/ENZYMES"
    assert config.data_set.__repr__() == "StandardTUDataset"


def test_config_cpu():
    config_filepath = "tests/config/classification_test_config_no_override_cpu.py"
    config = ConfigInitializer.from_file(file_name=config_filepath)
    assert config.dataset_name == "ENZYMES"
    assert config.num_classes == 6
    assert config.num_features == 3
    assert config.model == cgc_classification
    assert config.Evaluation.__repr__() == "Classification Evaluation Object"
    assert config.optimizer == "adams"
    assert config.net_dimension == 364
    assert config.learning_rate == 0.0005
    assert config.dataset_save_path == "tests/data/"
    assert config.test_ratio == 10
    assert config.batch_size == 64
    assert config.shuffle == True
    assert config.num_epoch == 2
    assert config.num_cv == 2
    assert config.device == torch.device("cpu")
    assert config.result_save_dir == "tests/results/ENZYMES"
    assert config.data_set.__repr__() == "StandardTUDataset"



"""
Evaulation object for classification
"""
from typing import Type

import torch
from carate.evaluation.base import Evaluation
from carate.load_data import DatasetObject


class ClassificationEvaluation(Evaluation):
    def __repr__(self) -> str:
        return "Classification Evaluation Object"



"""
Perform tests on the functionality of running models against
a regression dataset

:author: Julian M. Kleber
"""
from typing import Type
import logging

import torch

from carate.run import RunInitializer
import carate.models.cgc_regression as CGCR
from carate.evaluation.regression import RegressionEvaluation
from carate.load_data import StandardDatasetMoleculeNet, StandardDatasetTUDataset
from tests.utils import check_dir_paths, check_result_files

logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


def test_regression_multitaksing():
    check_dir_paths()
    config_filepath = "tests/config/regression_alchemy_test.py"
    run_title = "ALCHEMY_test"
    data_set_name = "alchemy_full"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    result_dir = f"tests/results/{run_title}"
    assert str(runner.data_set) == "StandardTUDataset"
    runner.run()  # takes instance attributes as parameters for the run() function

    check_result_files(
        result_dir=result_dir,
        data_set_name=data_set_name,
        run_title=run_title,
        override=True,
    )


def test_regression_override():
    check_dir_paths()
    config_filepath = "tests/config/regression_test_config_override.py"
    run_title = "ZINC_test"
    data_set_name = "ZINC_test"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    result_dir = f"tests/results/{run_title}"
    assert str(runner.data_set) == "StandardTUDataset"
    runner.run()  # takes instance attributes as parameters for the run() function

    check_result_files(
        result_dir=result_dir, run_title=run_title, data_set_name=data_set_name
    )


def test_regression_no_override():
    check_dir_paths()

    config_filepath = "tests/config/regression_test_config_no_override.py"
    run_title = "ZINC_test"
    data_set_name = "ZINC_test"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    result_dir = f"tests/results/{run_title}/"
    assert str(runner.data_set) == "StandardTUDataset"
    runner.run()  # takes instance attributes as parameters for the run() function

    check_result_files(
        result_dir=result_dir,
        run_title=run_title,
        data_set_name=data_set_name,
        override=True,
    )



import pandas as pd

from carate.plotting.base_plots import (
    plot_range_band_single,
    plot_range_band_multi,
    get_stacked_list,
)


def test_single_plot():
    path_to_directory = "./notebooks/data/"
    result = get_stacked_list(
        path_to_directory=path_to_directory,
        column_name="Acc",
        num_cv=5,
        json_name="ENZYMES.json",
    )
    plot_range_band_single(
        result, key_val="Acc_train", file_name="ENZYMES_accuracy", save_dir="./plots"
    )



import os
import shutil

import subprocess

from amarium.utils import search_subdirs

from tests.utils import check_result_files

def test_cli_func():
    if os.path.isdir("tests/data"):
        shutil.rmtree("tests/data")
    if os.path.isdir("tests/results"):
        shutil.rmtree("tests/results")
    subprocess.run(["bash", "install.sh"])
    config_filepath = "tests/config/regression_test_config_override.py"
    result = subprocess.run(
        ["carate", "-c", config_filepath],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    errs = result.stderr.decode()

    run_title = "ZINC_test"
    data_set_name = "ZINC_test"
    result_dir = f"tests/results/{run_title}"
    check_result_files(
        result_dir=result_dir, run_title=run_title, data_set_name=data_set_name
    )


"""
Tests about the classficiation abilities of the models

:author: Julian M. Kleber
"""
import os
import shutil

import torch

from amarium.utils import search_subdirs, load_json_from_file

from carate.run import RunInitializer
import carate.models.cgc_classification as CGCC
from carate.evaluation.classification import ClassificationEvaluation
from carate.load_data import StandardDatasetMoleculeNet, StandardDatasetTUDataset
from carate.config import Config


import logging

from typing import Type

logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


def test_classification_override():
    check_dir_paths()
    config_filepath = "tests/config/classification_test_config_override.py"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    runner.run()  # takes instance attributes as parameters for the run() function

    # check result files
    result_files, result_dirs = search_subdirs(dir_name="tests/results/ENZYMES/data")
    reference_dirs = [
        "tests/results/ENZYMES/data/CV_0",
        "tests/results/ENZYMES/data/CV_1",
    ]
    assert len(result_dirs) == 2
    for dir_name in result_dirs:
        assert dir_name in reference_dirs

    assert len(result_files) == 2
    reference_files = [
        "tests/results/ENZYMES/data/CV_0/ENZYMES.json",
        "tests/results/ENZYMES/data/CV_1/ENZYMES.json",
    ]
    for name in result_files:
        assert name in reference_files

    # check result checkpoints
    result_files, result_dirs = search_subdirs(
        dir_name="tests/results/ENZYMES/checkpoints"
    )

    reference_dirs = [
        "tests/results/ENZYMES/checkpoints/CV_0",
        "tests/results/ENZYMES/checkpoints/CV_1",
    ]
    assert len(result_dirs) == 2
    for dir_name in result_dirs:
        assert dir_name in reference_dirs

    assert len(result_files) == 2
    reference_files = [
        "tests/results/ENZYMES/checkpoints/CV_0/ENZYMES.tar",
        "tests/results/ENZYMES/checkpoints/CV_1/ENZYMES.tar",
    ]
    for name in result_files:
        assert name in reference_files

    result_dir_content = os.listdir("tests/results/ENZYMES/model_parameters")
    result_dir_content_data = [x for x in result_dir_content if x.endswith(".json")]
    assert len(result_dir_content_data) == 1
    assert "model_architecture.json" in result_dir_content_data


def test_classification_no_override():
    check_dir_paths()
    config_filepath = "tests/config/classification_test_config_no_override.py"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    runner.run()  # takes instance attributes as parameters for the run() function

    # check result files
    result_files, result_dirs = search_subdirs(dir_name="tests/results/ENZYMES/data")
    reference_dirs = [
        "tests/results/ENZYMES/data/CV_0",
        "tests/results/ENZYMES/data/CV_1",
    ]
    assert len(result_dirs) == 2
    for dir_name in result_dirs:
        assert dir_name in reference_dirs

    assert len(result_files) == 2
    reference_files = [
        "tests/results/ENZYMES/data/CV_0/ENZYMES.json",
        "tests/results/ENZYMES/data/CV_1/ENZYMES.json",
    ]
    result_cv1 = load_json_from_file("tests/results/ENZYMES/data/CV_1/ENZYMES.json")
    assert len(result_cv1["Loss"]) == 2
    for name in result_files:
        assert name in reference_files

    # check result checkpoints
    result_files, result_dirs = search_subdirs(
        dir_name="tests/results/ENZYMES/checkpoints"
    )

    reference_dirs = [
        "tests/results/ENZYMES/checkpoints/CV_0",
        "tests/results/ENZYMES/checkpoints/CV_1",
    ]

    assert len(result_dirs) == 2
    for dir_name in result_dirs:
        assert dir_name in reference_dirs

    assert len(result_files) == 4
    reference_files = [
        "tests/results/ENZYMES/checkpoints/CV_0/ENZYMES_Epoch-1.tar",
        "tests/results/ENZYMES/checkpoints/CV_0/ENZYMES_Epoch-2.tar",
        "tests/results/ENZYMES/checkpoints/CV_1/ENZYMES_Epoch-1.tar",
        "tests/results/ENZYMES/checkpoints/CV_1/ENZYMES_Epoch-2.tar",
    ]
    for name in result_files:
        assert name in reference_files

    result_dir_content = os.listdir("tests/results/ENZYMES/model_parameters")
    result_dir_content_data = [x for x in result_dir_content if x.endswith(".json")]
    assert len(result_dir_content_data) == 1
    assert "model_architecture.json" in result_dir_content_data


def check_dir_paths():
    if os.path.isdir("tests/data"):
        shutil.rmtree("tests/data")
    if os.path.isdir("tests/results"):
        shutil.rmtree("tests/results")



"""The runner object needs to be absolutely 
bullet proof in initialization. One false initialization
leads to weird results 

:author: Julian M. Kleber
"""
from carate.run import RunInitializer


def test_runner():
    # Regression
    config_filepath = "tests/config/regression_test_config_no_override.py"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    assert runner.dataset_name == "ZINC_test"
    assert (
        str(runner.Evaluation)
        == "Evaluation for cgc_regression with the Default evaluation"
    )
